"""
Contains an helper method to directly edit the xml
representation of the Presentation.
"""

from pptx.oxml.xmlchemy import OxmlElement


# https://github.com/scanny/python-pptx/issues/71
# https://groups.google.com/g/python-pptx/c/UTkdemIZICw
def SubElement(parent, tagname, before=(), **kwargs):
    """
    Adds the given tag, with attributes given in `kwargs`,
    inside `parent`, and as late as possible,
    but before all elements of tag in `before`.

    Returns the added element.
    """
    element = OxmlElement(tagname)
    element.attrib.update(kwargs)
    if before:
        parent.insert_element_before(element, *before)
    else:
        parent.append(element)
    return element



"""
Helper functions to interact with the Presentation.

In particular, contains functions to remove, move and create various elements.
"""

import logging
import os
import sys

import copy

from pptx import Presentation
from pptx.chart.data import CategoryChartData
from pptx.table import _Column, _Row

from .copy import textframe_copy


def export(pres, outpath):
    """
    Saves the given presentation at the given path, while deleting the previous file.
    """

    try:
        os.remove(outpath)
        logging.getLogger(__name__).info("Former %s deleted.", outpath)
    except OSError:
        pass
    pres.save(outpath)


def remove_slide(pres, idx):
    """
    Removes the slide at the given index from the presentation.
    """

    rId = pres.slides._sldIdLst[idx].rId
    pres.part.drop_rel(rId)
    del pres.slides._sldIdLst[idx]


def delete_run(r):
    """
    Deletes the given run from its paragraph.
    """
    r._r.getparent().remove(r._r)


def clear_slides(pres):
    """
    Deletes all slides from the presentation, in reverse order.
    """
    for idx in range(len(pres.slides) - 1, -1, -1):
        remove_slide(pres, idx)


def insert_slide(pres, layout, idx):
    """
    Creates a new slide with the given layout (either an actual Layout,
    or the name of a layout), and inserts it at index idx.
    """

    slide = append_slide(pres, layout)
    move_slide(pres.slides, slide, idx)
    return slide


def move_slide(slides, slide, new_idx):
    """
    Moves the given slide at the given index.
    """
    slides._sldIdLst.insert(new_idx, slides._sldIdLst[slides.index(slide)])


def fill_placeholders(slide):
    """
    Copy the content of the placeholders (actual PowerPoint placeholders,
    not placeholders as usually defined in this library) of the layout of
    the slide into the placeholders of the slide.
    """

    layout = slide.slide_layout
    for ph in layout.placeholders:
        if ph.has_text_frame:
            try:
                phid = ph.placeholder_format.idx
                tf = slide.placeholders[phid].text_frame
            except KeyError:
                continue
            textframe_copy(ph.text_frame, tf)


def append_slide(pres, layout):
    """
    Adds a slide with the given layout (either an actual layout or the name
    of an existing layout) at the end of the presentation.
    """

    if isinstance(layout, str):
        nl = pres.slide_masters[0].slide_layouts.get_by_name(layout)
        if nl is None:
            logging.getLogger(__name__).error(
                "append_slide: layout \'%s\' inexistant", layout)
            sys.exit(1)
        layout = nl

    slide = pres.slides.add_slide(layout)
    fill_placeholders(slide)
    return slide


def etalonnage(prespath, outpath):
    """
    Loads the presentation at the given path,
    then for each shape write the shape index inside,
    and lastly export the produced presentation.
    """
    pres = Presentation(prespath)
    for idx, slide in enumerate(pres.slides):
        logging.getLogger(__name__).info("Slide %d has slide_id %s", idx,
                                         slide.slide_id)

    clear_slides(pres)
    for layout in pres.slide_masters[0].slide_layouts:
        slide = pres.slides.add_slide(layout)
        for ph in slide.placeholders:
            ph.text = f"placeholder {ph.placeholder_format.idx}"
    export(pres, outpath)


def chart_replace_data(chart, values):
    """
    Updates the data in the chart.
    """

    chart_data = CategoryChartData()
    chart_data.categories = chart.plots[0].categories
    chart_data.add_series('Values', values)
    chart.replace_data(chart_data)


def find_shape(slide, name):
    """
    Finds the first shape in the slide with the given name.

    Convenient to access a shape by its name.
    """

    for shape in slide.shapes:
        if shape.name == name:
            return shape
    return None


def table_dup_row(table, idx, n=1, to=None):
    """
    Duplicates a table row, in particular borders and text are copied.
    - idx is the index of the row we want to copy
    - n is the number of times we want to copy the row
    - to is the index where we want to insert the row
    """

    if to is None:
        to = idx
    to = (to + len(table.rows)) % len(table.rows)

    new_row = table._tbl.tr_lst[idx]
    num = -1
    for rid, rxml in enumerate(table._tbl.iterchildren()):
        if rxml.tag == new_row.tag:
            num += 1
            if num == to:
                break

    for _ in range(n):
        new_row = copy.deepcopy(new_row)
        table._tbl.insert(rid, new_row)

    if n == 1:
        return _Row(new_row, table)
    return None


def table_dup_column(table, idx, n=1, to=None, keep_width=False):
    """
    Duplicates a column of the table, in particular borders and text are copied.
    - idx is the index of the column we want to copy
    - n is the number of times we want to copy the column
    - to is the index where we want to insert the column
    - keep_width indicates whether we want to shrink columns to add the new ones
    so that the total width is the same, or if columns are simply added.
    """

    total_width = table._graphic_frame.width
    colwidth = table.columns[idx].width
    mult = total_width / (total_width + n * colwidth)

    if to is None:
        to = idx
    to = (to + len(table.columns)) % len(table.columns)

    new_col = table._tbl.tblGrid.gridCol_lst[idx]
    num = -1
    for cid, cxml in enumerate(table._tbl.tblGrid.iterchildren()):
        if cxml.tag == new_col.tag:
            num += 1
            if num == to:
                break
    for _ in range(n):
        new_col = copy.deepcopy(new_col)
        table._tbl.tblGrid.insert(cid, new_col)

    for tr in table._tbl.tr_lst:
        new_cell = tr.tc_lst[idx]
        num = -1
        for cid, cxml in enumerate(tr.iterchildren()):
            if cxml.tag == new_cell.tag:
                num += 1
                if num == to:
                    break
        for _ in range(n):
            new_cell = copy.deepcopy(new_cell)
            tr.insert(cid, new_cell)

    if keep_width:
        for col in table.columns:
            col.width = int(col.width * mult)

    if n == 1:
        return _Column(new_col, table)
    return None



"""
Contains functions to easily edit the style of an element.
- Cell borders
- Table borders
- Font
- Paragraph
- TextFrame
"""

from .xml import SubElement
from pptx.dml.color import RGBColor as RGB
from pptx.enum.dml import MSO_LINE
from pptx.util import Pt


def set_cell_borders(cell,
                     border_color=RGB(0, 0, 0),
                     border_width=Pt(1),
                     border_style=MSO_LINE.DASH,
                     left=False,
                     top=False,
                     right=False,
                     bot=False):
    """
    The former border is kept (when adding a border, the former is removed if it exists)
    - border_color is the color of the border
    - border_width is the width of the border
    - border_style is the style of the border
    - left, top, right, bot indicate which borders are to be set
    """

    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()

    borders = [('L', left), ('R', right), ('T', top), ('B', bot)]
    for idx, (side, do) in enumerate(borders):
        if not do:
            continue
        tag = f"a:ln{side}"

        # Remove previous border if it exists
        tcPr.remove_all(tag)

        # Borders have to be declared in a specific order (L, R, T, B)
        # or they're not displayed
        before = [f"a:ln{s}" for (s, _) in borders[idx + 1:]]
        # borders must appear in the xml before the cell style
        # or they're not displayed
        before += ['a:solidFill', 'a:noFill']

        ln = SubElement(tcPr,
                        tag,
                        before=before,
                        w=str(border_width),
                        cap='flat',
                        cmpd='sng',
                        algn='ctr')

        solidFill = SubElement(ln, 'a:solidFill')
        # rgb color:
        colorval = "".join(map(lambda v: f"{v:02X}", border_color))
        SubElement(solidFill, 'a:srgbClr', val=colorval)
        # predefined color:
        # SubElement(solidFill, 'a:schemeClr', val="tx1")

        SubElement(ln, 'a:prstDash', val=MSO_LINE.to_xml(border_style))
        SubElement(ln, 'a:round')
        SubElement(ln, 'a:headEnd', type='none', w='med', len='med')
        SubElement(ln, 'a:tailEnd', type='none', w='med', len='med')


def set_rect_borders(table,
                     left,
                     top,
                     right,
                     bot,
                     outer=False,
                     inner=False,
                     **kwargs):
    """
    - left, top, right, bot are the coordinates of the border cells
    - outer indicates whether to set outer borders
    - inner indicates whether to set inner borders
    - kwargs are propagated to the Cell borders
    """

    assert left >= 0
    assert top >= 0
    assert right < len(table.columns)
    assert bot < len(table.rows)
    assert left <= right
    assert top <= bot

    for x in range(left, right + 1):
        for y in range(top, bot + 1):
            topb = outer if y == top else inner
            botb = outer if y == bot else inner
            leftb = outer if x == left else inner
            rightb = outer if x == right else inner

            set_cell_borders(table.cell(y, x),
                             **kwargs,
                             left=leftb,
                             top=topb,
                             right=rightb,
                             bot=botb)

    if not outer:
        return

    # PowerPoint has a bug (?), for cells strictly inside an array
    # only right and bottom borders are looked at, so we need to add
    # borders on cells outside of the selection

    for y in range(top, bot + 1):
        # right border of the left column
        if left != 0:
            set_cell_borders(table.cell(y, left - 1), **kwargs, right=True)
        # left border of the right column
        if right != len(table.columns) - 1:
            set_cell_borders(table.cell(y, right + 1), **kwargs, left=True)

    for x in range(left, right + 1):
        # lower border of upper row
        if top != 0:
            set_cell_borders(table.cell(top - 1, x), **kwargs, bot=True)
        # upper border of lower row
        if bot != len(table.rows) - 1:
            set_cell_borders(table.cell(bot + 1, x), **kwargs, top=True)


def font_style(ft,
               size=None,
               font=None,
               bold=None,
               italic=None,
               font_rgb=None,
               font_brightness=None,
               underline=None,
               language=None):
    """
    - size is the font size, usually Pt or Cm:
    https://python-pptx.readthedocs.io/en/latest/api/util.html
    - font is the name of the font (string, case sensitive)
    - bold indicates whether the text is bold
    - italic indicates whether the text is italic
    - font_rgb is the color of the text, of type RGBColor, built either from 3 integers
    or from a string corresponding to 3 hexadecimal integers:
    https://python-pptx.readthedocs.io/en/latest/api/dml.html#pptx.dml.color.RGBColor
    - font_brightness is the brightness of the font, a float between -1 and 1
    - underline indicates whether the text is underlined
    - language indicates the language of the text:
    https://python-pptx.readthedocs.io/en/latest/api/enum/MsoLanguageId.html
    """

    if size is not None:
        ft.size = size
    if font is not None:
        ft.name = font
    if bold is not None:
        ft.bold = bold
    if italic is not None:
        ft.italic = italic
    if font_rgb is not None:
        ft.color.rgb = font_rgb
    if font_brightness is not None:
        ft.color.brightness = font_brightness
    if underline is not None:
        ft.underline = underline
    if language is not None:
        ft.language_id = language


def text_style(tf, halign=None, **kwargs):
    """
    - halign indicates how the text is horizontally aligned
    https://python-pptx.readthedocs.io/en/latest/api/enum/PpParagraphAlignment.html
    - kwargs are propagated to the Paragraph
    """

    for paragraph in tf.paragraphs:
        font_style(paragraph.font, **kwargs)
        if halign is not None:
            paragraph.alignment = halign


def cell_style(cell,
               valign=None,
               back_rgb=None,
               back_brightness=None,
               margin_bottom=None,
               margin_top=None,
               margin_left=None,
               margin_right=None,
               **kwargs):
    """
    - valign indicated how the text is vertically aligned
    https://python-pptx.readthedocs.io/en/latest/api/enum/MsoVerticalAnchor.html
    - back_rgb is the background color of the cell
    https://python-pptx.readthedocs.io/en/latest/api/dml.html#pptx.dml.color.RGBColor
    - back_brightness is the background brightness of the cell, a float between -1 and 1
    - margin_top, margin_left, margin_bottom, margin_right are the margins of the cell
    https://python-pptx.readthedocs.io/en/latest/api/util.html
    - kwargs are propagated to the TextFrame
    """

    if margin_bottom is not None:
        cell.margin_bottom = margin_bottom
    if margin_top is not None:
        cell.margin_top = margin_top
    if margin_right is not None:
        cell.margin_right = margin_right
    if margin_left is not None:
        cell.margin_left = margin_left
    if valign is not None:
        cell.vertical_anchor = valign
    if back_brightness is not None:
        cell.fill.solid()
        cell.fill.fore_color.brightness = back_brightness
    if back_rgb is not None:
        if not back_rgb:
            cell.fill.background()
        else:
            cell.fill.solid()
            cell.fill.fore_color.rgb = back_rgb

    text_style(cell.text_frame, **kwargs)


def paragraph_append(paragraph, txt, **kwargs):
    """
    Adds a run to the paragraph, with the given text.

    - kwargs are propagated to the Run.
    """
    run = paragraph.add_run()
    run.text = txt
    font_style(run.font, **kwargs)



"""
Everything related to placeholders and text or image replacement.
"""

import logging
import re
import traceback

from PIL import Image

from .presmanager import delete_run

# Define the characters marking the opening of a placeholder.
CMD_P1_ = CMD_P2_ = '{'

# Define the characters marking the closing of a placeholder.
CMD_S1_ = CMD_S2_ = '}'

# Characters which are replaced before executing the code.
# By default, the only replaced characters are ‘ and ’ which are replaced by a straight single quote.
REPLACE_CHARS_ = {'‘': '\'', '’': '\''}


def cmd_p():
    """
    Returns the opening string of a placeholder.
    """
    return f"{CMD_P1_}{CMD_P2_}"


def cmd_s():
    """
    Returns the closing string of a placeholder.
    """
    return f"{CMD_S1_}{CMD_S2_}"


def cmd_pattern():
    """
    Returns the (compiled) pattern corresponding to a text placeholder.
    """
    return re.compile(f"{cmd_p()}(.*?){cmd_s()}")


def img_pattern():
    """
    Returns the (compiled) pattern corresponding to an image placeholder.
    """
    return f"{cmd_p()}img:(.*?):(.*?){cmd_s()}"


def interp(cmd, **kwargs):
    """
    Given python code extracted from a placeholder, interpret it with kwargs
    as environment, and return the value fetched from the code as a string.

    Characters in cmd are replaced if they appear in REPLACE_CHARS_, then
    the command is executed with
    ```py
    exec(f"_ret={cmd}", None, kwargs)
    ```
    and lastly the value of `_ret` is fetched and returned as a string.

    In case of error, an empty string is returned.

    Using exec is of course very unsafe, but we assume the users knows what
    he is executing, or else that he at least knows the risks.
    """

    if '_r' in kwargs and '_c' in kwargs:
        context = f"[Slide {kwargs['_idx']}, cell at row {kwargs['_r']} and col {kwargs['_c']}]"
    else:
        context = f"[Slide {kwargs['_idx']}]"

    logging.getLogger(__name__).debug("%s Interpreting command '%s'", context,
                                      cmd)
    for cfrom, cto in REPLACE_CHARS_.items():
        cmd = cmd.replace(cfrom, cto)
    try:
        exec(f"_ret={cmd}", None, kwargs)
    except:
        logging.getLogger(__name__).warning("%s Error executing:\n%s\n%s",
                                            context, cmd,
                                            traceback.format_exc())
        # raise again or just keep going ?
        # for now just keep going
        return ''
    return kwargs['_ret']


def run_replace(run, **kwargs):
    """
    Find and replace placeholders inside a run, with the given local environment.

    Placeholders are replaced in order.

    Placeholders are removed before executing the code.
    """

    cmd_pat = cmd_pattern()
    # do not use 'subn' since we want to remove the code before executing
    m = cmd_pat.search(run.text)
    while m is not None:
        # delete code, then interpret and insert where the code used to be
        # prevents issues when the piece of code duplicates the current slide
        run.text = run.text[:m.start()] + run.text[m.end():]
        val = str(interp(m.group(1), **kwargs))
        run.text = run.text[:m.start()] + val + run.text[m.start():]

        m = cmd_pat.search(run.text)


def cell_replace(cell, _cell=None, **kwargs):
    """
    Find and replace placeholders inside a cell,
    with the given local environment.

    Adds the cell to the environment, as `_cell`.
    """
    textframe_replace(cell.text_frame, **kwargs, _cell=cell)


def paragraph_replace(par, _p=None, **kwargs):
    """
    Find and replace placeholders inside a paragraph,
    with the given local environment.

    Adds the paragraph to the environment, as `_p`.

    The order in which placeholders are interpreted is first by considering
    runs independently, in text order, and then checking whether the paragraph
    has placeholders over run boundaries.

    Placeholders are fully removed before being interpreted, and the result of
    the execution is inserted in the first run.
    """

    for run in par.runs:
        run_replace(run, **kwargs, _p=par)

    idx = 0
    runs = list(par.runs)
    matches = list(cmd_pattern().finditer(par.text))
    for match in matches:
        idcs = -1  # char start
        while idcs == -1:
            rt = runs[idx].text
            if not rt:
                continue
            idcs = rt.find(cmd_p())

            if idcs == -1 and rt[-1] == CMD_P1_ and \
               runs[idx + 1].text[0] == CMD_P2_:
                idcs = len(rt) - 1

            idx += 1

        idrs = idx - 1  # run start

        idce = 1  # char end
        while idce == 1:
            rt = runs[idx].text
            if not rt:
                continue
            idce = rt.find(cmd_s()) + 2

            if idce == 1 and rt[-1] == CMD_S1_ and \
               runs[idx + 1].text[0] == CMD_S2_:
                idx += 2
                idce = 1
                break

            idx += 1

        idre = idx - 1  # run end

        runs[idrs].text = runs[idrs].text[:idcs]
        runs[idre].text = runs[idre].text[idce:]

        idx = idrs + 1

        for i in range(idrs + 1, idre):
            delete_run(runs[i])
        if not runs[idre].text:
            delete_run(runs[idre])

        # delete code, then interpret and insert where the code used to be
        # prevents issues when the piece of code duplicates the current slide
        val = interp(match.group(1), **kwargs, _p=par)
        runs[idrs].text += str(val)

        if not runs[idrs].text:
            delete_run(runs[idrs])
        runs = list(par.runs)


def textframe_replace(tf, _tf=None, **kwargs):
    """
    Find and replace placeholders inside a TextFrame,
    with the given local environment.

    Adds the TextFrame to the environment, as `_tf`.

    If the TextFrame contains an image placeholder, the container is removed
    and the image added at its position. Otherwise, paragraphs are interpreted
    in text order separately (textual placeholders are not searched across
    paragraph boundaries).
    """

    # match image
    m = re.search(img_pattern(), tf.text)

    if m is None:
        for par in tf.paragraphs:
            paragraph_replace(par, **kwargs, _tf=tf)
        return
        # don't replace across paragraphs

    # insert the image
    slide = tf.part.slide
    sh = tf._element.getparent()
    l, t, w, h = sh.x, sh.y, sh.cx, sh.cy

    modifs = set(m.group(1))
    file = interp(m.group(2), **kwargs, _tf=tf)
    slide.shapes.element.remove(sh)

    if not file:
        # ignore placeholders with undefined image
        return

    context = f"[Slide {kwargs['_idx']}] textframe_replace (image '{file}', modifiers={modifs})"

    try:
        with Image.open(file, mode='r') as im:
            width = im.width
            height = im.height
    except (FileNotFoundError, PermissionError, OSError):
        logging.getLogger(__name__).exception("%s: ", context)
        return

    hm = {'r', 'l', 'c'}
    vm = {'t', 'b', 'm'}
    allm = hm | vm

    if not modifs.issubset(allm):
        logging.getLogger(__name__).info("%s: unknown modifiers %s ignored",
                                         context, modifs - allm)

    if len(modifs.intersection(hm)) > 1:
        logging.getLogger(__name__).warning(
            "%s: several horizontal modifiers %s", context,
            modifs.intersection(hm))
        return
    if len(modifs.intersection(vm)) > 1:
        logging.getLogger(__name__).warning(
            "%s: several vertical modifiers %s", context,
            modifs.intersection(vm))
        return

    factor = min(w / width, h / height)
    if modifs.isdisjoint(hm):
        width = w
    else:
        # scaling horizontal
        width *= factor
    if modifs.isdisjoint(vm):
        height = h
    else:
        # scaling vertical
        height *= factor

    if 'r' in modifs:
        # aligned right
        l += w - width
    if 'c' in modifs:
        # centered (horizontal)
        l += (w - width) // 2
    if 'b' in modifs:
        # aligned bottom
        t += h - height
    if 'm' in modifs:
        # centered (vertically)
        t += (h - height) // 2

    slide.shapes.add_picture(file, l, t, width=width, height=height)


def row_replace(row, _c=None, **kwargs):
    """
    Find and replace placeholders inside a table row,
    with the given local environment.

    Adds the table column index to the environment, as `_c`.

    Cells are interpreted from left to right.
    """

    # I did not use enumerate to avoid issues if interpreting changes the number of rows
    # but not sure it's much better using range(len(...))
    for _c in range(len(row.cells)):
        cell_replace(row.cells[_c], **kwargs, _c=_c)


def table_replace(table, _r=None, _table=None, **kwargs):
    """
    Find and replace placeholders inside a table,
    with the given local environment.

    Adds the table and row index to the environment, as `_table` and `r`.

    Rows are interpreted from top to bottom.
    """

    # I did not use enumerate to avoid issues if interpreting changes the number of rows
    # but not sure it's much better using range(len(...))
    for _r in range(len(table.rows)):
        row_replace(table.rows[_r], **kwargs, _table=table, _r=_r)


def slide_replace(slide, _sl=None, **kwargs):
    """
    Find and replace placeholders inside a slide,
    with the given local environment.

    Adds the slide in the environment, as `_sl`.

    Shapes are interpreted from in the order defined in the slide, which
    can be seen and edited by clicking on the `Select` button in the `Home`
    tab and then `Selection Pane` (bottom-up).

    Only shape TextFrames and Tables are interpreted.
    """

    for sh in slide.shapes:
        if sh.has_text_frame:
            textframe_replace(sh.text_frame, **kwargs, _sl=slide)
        if sh.has_table:
            table_replace(sh.table, **kwargs, _sl=slide)


def pres_replace(pres, _idx=None, _pres=None, **kwargs):
    """
    Find and replace placeholders inside the presentation,
    with the given local environment.

    Adds the Presentation and slide index in the environment,
    as `_pres` and `_idx`.

    Slides are interpreted in ascending order.
    """

    for idx, slide in enumerate(pres.slides):
        slide_replace(slide, **kwargs, _pres=pres, _idx=idx)
